/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of cryptoFTP - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU General Public License (see COPYING.GPL for details).
 *
 * $Id: parser.c,v 1.6 2000/05/05 22:20:25 malekith Exp $
 */

#include "config.h"
#include "cftp.h"
#include "utils.h"
#include "cipher.h"

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <stdio.h>
#include <unistd.h>


void close_link();
void create_link();

void cdie(const char *str);
extern char *user, *host;
extern int port;

void compl_free();
void compl_add(const char *fname);

extern struct clib_state *ctx;

int split_hp(char *p)
{
	if (strchr(p, '^')) {
		*strchr(p, '^') = 0;
		return atoi(strchr(host, 0) + 1);
	} else if (strchr(host, ':')
#ifdef IPv6
		&& !strchr(strchr(host, ':') + 1, ':')
#endif
	) {
		*strchr(host, ':') = 0;
		return atoi(strchr(host, 0) + 1);
	} else
		return 0;
}

static void c_close(char *p)
{
	(void)p;
	close_link();
}

static void c_bye(char *p)
{
	c_close(p);
	printf("<<<--- - - bye bye === : - )\n");
	exit(0);
}

static void c_die(char *p)
{
	(void)p;
	printf("!!!! dying w/o closing sockets");
	exit(1);
}

extern const char *cipher_to_use;
static void c_cipher(char *s)
{
	struct cipher2 *c;
	char *n;

	n = s;
	for (; *s && *s != ' '; s++)
		if (*s >= 'A' && *s <= 'Z')
			*s += 0x20;
	*s = 0;

	if (*n) {
		c = find_cipher(n);
		if (!c)
			printf("!> cipher '%s' not found\n", n);
		else if (c->disabled)
			printf("!> cipher '%s' is disabled: %s\n", c->name,
			       c->disabled);
		else
			printf("%% will use %s on next connection\n",
			       cipher_to_use = c->name);
	} else {
		printf("cipher for next connection: %s\n", cipher_to_use);
	}
}


static void c_clist(char *s)
{
	int i;
	(void)s;
	printf("%10s Key Block %s\n", "Name", "Disabled");
	for (i=0; ciphers[i].name; i++) {
		if (ciphers[i].creator) {
			struct cipher *t = ciphers[i].creator();
			printf("%10s %3d %3d   %s\n", ciphers[i].name, 
				t->keylen, t->blocklen, ciphers[i].disabled ?
				ciphers[i].disabled : "nope");
			t->free(t);
		} else
			printf("%10s ??? ???   %s\n", ciphers[i].name, 
				ciphers[i].disabled ? ciphers[i].disabled : 
					"CREATOR UNKNOWN");
	}
}

static void c_open(char *p)
{
	if (strchr(p, '@')) {
		char *hp = p;
		
		while (*hp != '@')
			hp++;
		*hp++ = 0;
		xfree(user);
		user = xstrdup(p);
		xfree(host);
		host = xstrdup(hp);
	} else {
		xfree(host);
		host = xstrdup(p);
	}
	
	port = split_hp(host);
	
	create_link();
}


static void c_info(char *p)
{
	p = clib_info(ctx);
	if (p)
		printf("INFO:\n%s\n", p);
	xfree(p);
}

static void c_pwd(char *p)
{
	p = clib_pwd(ctx);
	if (p)
		printf("PWD: \"%s\"\n", p);
	xfree(p);
}


void get_file(const char *name, const char *save_as, int compress);
void send_file(const char *name, const char *send_as, int compress);

static char *basename(char *p)
{
	return strrchr(p, '/') ?  strrchr(p, '/') + 1 : p;
}

static void c_put(char *p)
{
	send_file(p, basename(p), 0);
}

static void c_zput(char *p)
{
	send_file(p, basename(p), 1);
}

static void c_get(char *p)
{
	get_file(p, basename(p), 0);
}

static void c_zget(char *p)
{
	get_file(p, basename(p), 1);
}

static void display_file(struct clib_stat *buf)
{
	struct tm *t;
	static int year = 0;
	static char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	
	if (!year) {
		/* yeah... we're assuming noone will run this */
		/* program longer than a year ... */
		time_t now;
		
		time(&now);
		t = localtime(&now);
		year = t->tm_year;
	}
	
	t = localtime((time_t*)&buf->mtime);
	
	printf("%s %3d %-8s %-8s %8d %3.3s %2d ",buf->text_mode, buf->nlink, buf->user, buf->group, 
			buf->size, months[t->tm_mon], t->tm_mday);
	
	if (year == t->tm_year)
		printf("%02d:%02d", t->tm_hour, t->tm_min);
	else
		printf(" %d", t->tm_year + 1900);
	
	printf(" %s", buf->name);
	if (buf->symlink[0])
		printf(" -> %s", buf->symlink);
	printf("\n");
}

char *files[1000];
int fk;

char *make_name(struct clib_stat *buf)
{
	char *p;
	p = xmalloc0(strlen(buf->name) + 2);
	strcpy(p, buf->name);
	if (buf->isdir)
		*strchr(p, 0) = '/';
	else if (buf->symlink[0])
		*strchr(p, 0) = '@';
	return p;
}

void display_names()
{
	int i,len,maxlen,col,ncol;
        int columns = 79;

        if (getenv("COLUMNS"))
                columns = atoi(getenv("COLUMNS"));

        if (columns < 30)
                columns = 30;

        if (columns > 300)
                columns = 300;
	
	if (!fk)
		return;
		
	for (len = 0, maxlen = 0, i = 0; i < fk; i++) {
		int l = strlen(files[i]);
		len += l + 2;
		if (maxlen < l)
			maxlen = l;
	}
	
	if (len < columns - 1) {
		for (len = 0, i = 0; i < fk; i++)
			printf("%s  ", files[i]);
		printf("\n");
		goto bye;
	}
	
	maxlen++;
	ncol = columns / maxlen;
	len = columns / ncol;
	
	for (i = 0, col = 0; i < fk; i++) {
		printf("%-*s", len, files[i]);
		col++;
		if (col >= ncol) { printf("\n"); col = 0; }
	}
	if (col)
		printf("\n");
bye:
	for (i = 0; i < fk; i++)
		xfree(files[i]);
}

static void do_ls(int print)
{
	struct clib_stat buf;
	
	compl_free();
	fk = 0;
	
	while (clib_ls(ctx, &buf) == 0) {
		compl_add(buf.name);
		
		if ((print & 4) && buf.name[0] == '.')
			continue;
			
		if (print & 1)
			display_file(&buf);
		else if (print & 2)
			files[fk++] = make_name(&buf);
	}
	
	if (print & 2)
		display_names();
}

int forward(struct clib_state *ctx, char *p);

static void c_fwd(char *p)
{
	forward(ctx, p);
}

int anonfwd(char *p);

static void c_anonfwd(char *p)
{
	anonfwd(p);
}

static void c_ls(char *p)
{
	int fl;
	
	if (strchr(p, 'l'))
		fl = 1;
	else 
		fl = 2;
	
	if (!strchr(p, 'a'))
		fl |= 4;
	
	do_ls(fl);
}

static void c_mkdir(char *p) { clib_mkdir(ctx, p); }
static void c_rmdir(char *p) { clib_rmdir(ctx, p); }
static void c_cd(char *p) { if (clib_cd(ctx, p) == 0) compl_free(); }

static void c_rm(char *p)
{
	if (strncmp(p, "-r ", 3) == 0) {
		p += 3;
		while (isspace(*p))
			p++;
		clib_rmr(ctx, p);
	} else
		clib_rm(ctx, p);
}

/*
#include <fcntl.h>
#include <unistd.h>*/
/*void c_test(char*p){int fd;byte buf[100000];random_get(buf,sizeof buf);
fd=open(p,O_WRONLY|O_CREAT,0666);write(fd,buf,sizeof buf);close(fd);}*/

struct command {
	const char *name;
	void (*executor)(char *);
	int flags;
	const char *args;
	const char *help;
};

#define REQ_CTX		0x00000001
#define NO_PARAM	0x00000002
#define REQ_PARAM	0x00000004

static void c_help(char *p);

static struct command cmds[] = {
{ "close", c_close, REQ_CTX|NO_PARAM, 0,
	"close a connection" },
{ "bye", c_bye, NO_PARAM, 0,
	"exit client" },
{ "die", c_die, NO_PARAM, 0,
	"exit client w/o closing any sockets" },
{ "cipher", c_cipher, 0, "[cipher]",
	"set cipher for next connection or display current one" },
{ "clist", c_clist, NO_PARAM, 0,
	"list available ciphers" },
{ "open", c_open, REQ_PARAM, "[user@]host[:port]",
	"open connection closing possible previous one" },
{ "info", c_info, NO_PARAM|REQ_CTX, 0,
	"display some info on cryptoFTP, as obtained from server" },
{ "pwd", c_pwd, NO_PARAM|REQ_CTX, 0,
	"print current working directory, as obtained from server" },
{ "get", c_get, REQ_PARAM|REQ_CTX, "filename",
	"get specified file from server, it's is stored in current local\n"
	"directory, with all leading path elements removed" },
{ "zget", c_zget, REQ_PARAM|REQ_CTX, "filename",
	"same as GET, but also reqests deflate compression in transfer" },
{ "put", c_put, REQ_PARAM|REQ_CTX, "filename",
	"send specified file to server, it's is stored in current remote\n"
	"directory (use PWD to check it), with all leading path elements\n"
	"removed" },
{ "zput", c_zput, REQ_PARAM|REQ_CTX, "filename",
	"same as PUT, but also uses deflate compression in transfer (if\n"
	"supports it)." },
{ "fwd", c_fwd, REQ_PARAM|REQ_CTX, "[remote_host:]remote_port[:local_port]",
	"ask server to forward REMOTE_HOST:REMOTE_PORT. local socket is\n" 
	"created and binded to LOCAL_PORT. LOCAL_PORT defaults to\n"
	"REMOTE_PORT and REMOTE_HOST defalts to ``localhost'' (i.e. server's\n"
	"``localhost''). all forwarding precesses are killed when bye is\n"
	"executed, but close command doesn't kill them." },
{ "ls", c_ls, REQ_CTX, "[-l] [-a]",
	"lists current remote directory. other dirs can't be listed." },
{ "mkdir", c_mkdir, REQ_PARAM|REQ_CTX, "directory_name",
	"make remote directory. it's created with permissions of 0777\n"
	"modified with server's umask, which is probably 022, which gives\n"
	"0755." },
{ "rmdir", c_rmdir, REQ_PARAM|REQ_CTX, "directory_name",
	"remove remote dierctory." },
{ "rm", c_rm, REQ_PARAM|REQ_CTX, "[-r] filename",
	"removes specified file or directory, recursing it if -r is given.\n"
	"recursion is done on server side, so it's damn fast and danger." },
{ "cd", c_cd, REQ_PARAM|REQ_CTX, "directory_name",
	"change working directory, server should understand ~/dir notation" },
{ "help", c_help, REQ_PARAM, "[topic]",
	"gives help on given topic. use ``help command_name'' to get help\n"
	"about given command, use ``help anaything_else'' to get list of\n"
	"commands" },
{ "anonfwd", c_anonfwd, REQ_PARAM, "host[:port] [remote_host:]remote_port[:local_port]",
	"starts up anonymous forwarding to REMOTE_HOST:REMOTE_PORT thru\n"
	"server HOST (at port PORT). local socket is binded to LOCAL_PORT\n"
	"LOCAL_PORT defaults to REMOTE_PORT and REMOTE_HOST defalts to\n"
	"``localhost'' (i.e. server's\n ``localhost''). all forwarding\n"
	"precesses are killed when ``bye'' command is executed, but\n"
	"``close'' command doesn't kill them." },
{ 0, 0, 0, 0, 0 }
};

static void c_help(char *p)
{
	struct command *c;
	
	for (c = cmds; c->name; c++)
		if (xstrcasecmp(c->name, p) == 0)
			break;
				
	if (c->name) { // [m
		printf("[32msyntax[0;2m: [1m%s[0m %s\n[32mdescription[0;2m:[0m\n%s\n",
			c->name, c->args?c->args:"", c->help);
		printf("[32mflags[0;2m: ");
		if (c->flags&REQ_CTX)
			printf("[[1mrequire connection[2m] ");
		if (c->flags&REQ_PARAM)
			printf("[[1mrequire parameter[2m] ");
		if (c->flags&NO_PARAM)
			printf("[[1mcan't get parameter[2m] ");
		printf("[0m\n");
	} else {
		printf("[31mno help found on [34;1m%s\n[32mavailable topics[0;2m:[0m\n", p);
		for (c = cmds; c->name; c++)
			printf("%s ", c->name);
		printf("\n");
	}
}

#ifdef COMPLETION
char *compl_filenames[4096];
int compl_index;
void compl_free()
{
	int i;
	for (i = 0; compl_filenames[i]; i++) {
		xfree(compl_filenames[i]);
		compl_filenames[i] = 0;
	}
	compl_index = 0;
}

void compl_add(const char *fname)
{
	if (compl_index < 4000)
		compl_filenames[compl_index++] = xstrdup(fname);
}

static void do_ls(int print);
void rl_on_new_line();

char *compl_cmd(char *text, int state)
{
	static int i, l;
	if (state == 0) {
		i = 0;
		l = strlen(text);
	}
		
	while (cmds[i].name)
		if (strncmp(text, cmds[i++].name, l) == 0)
			return xstrdup(cmds[i-1].name);
	return 0;
}

char *compl_gen(char *text, int state)
{
	static int i, l;
	if (state == 0) {
		i = 0;
		l = strlen(text);
	}
		
	if (compl_index == 0 && ctx) {
		printf("<--\nsending ls request to allow completion\n");
		do_ls(0);
		rl_on_new_line();
	}
	
	while (compl_filenames[i])
		if (strncmp(text, compl_filenames[i++], l) == 0)
			return xstrdup(compl_filenames[i-1]);
	return 0;
}
#else
void compl_free()
{
}
void compl_add(const char *fname)
{
	(void)fname;
}
#endif

void do_command(char *line, char *p)
{
	struct command *c;
	
	for (c = cmds; c->name; c++)
		if (xstrcasecmp(c->name, line) == 0)
			break;
	if (!c->name) {
		printf("!> no, no bad magic word: %s\n", line);
		return;
	}
	
	if ((c->flags&REQ_CTX) && ctx == 0)
		printf("!> sorry, this command requires valid connection\n");
	else if ((c->flags&REQ_PARAM) && *p == 0) {
		printf("!> no parameter given\n");
		c_help((char*)c->name);
	} else if ((c->flags&NO_PARAM) && *p) {
		printf("!> this command can't get parameter\n");
		c_help((char*)c->name);
	} else
		c->executor(p);
	
	if (ctx && ctx->error)
		printf("!> cftp error: %s\n", ctx->error);
}
