/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of libcftp - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License (see COPYING.LGPL 
 * for details).
 *
 * $Id: rsa.h,v 1.1.1.1 2000/04/22 10:15:53 malekith Exp $
 */

#ifndef _RSA_H
#define _RSA_H

#include "gmp.h"

void xmpz_to_byte(mpz_t d, byte *buf, int len);
void xmpz_from_byte(mpz_t d, byte *buf, int len);
void xmpz_get_rand(mpz_t v, int bits);
void xmpz_get_prime(mpz_t v, int bits);
void xmpz_get_rel_prime(mpz_t out, mpz_t p, int bits);

struct rsa_key {
	int bits;
	mpz_t n,e,d;
};

void rsa_encrypt(mpz_t out, mpz_t in, struct rsa_key *key);
void rsa_decrypt(mpz_t out, mpz_t in, struct rsa_key *key);
void rsa_destroy_key(struct rsa_key *key);
void rsa_generate_key(struct rsa_key *key, int bits);
void rsa_set_memory_alloc(void);
int rsa_read_pub_key(struct rsa_key *pub, const char *data);

#endif /* _RSA_H */

