/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of cryptoFTP - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU General Public License (see COPYING.GPL for details).
 *
 * $Id: config.c,v 1.1 2000/04/23 16:39:03 malekith Exp $
 */
#include "config.h"
#include "server.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct conf_var {
	struct conf_var *next;
	char *name;
	char *val;
};

static struct conf_var *conf_head;

static int convert(const char *string, struct conf_map *map, int def)
{
	while (isspace(*string))
		string++;
		
	for (; map->name; map++)
		if (xstrcasecmp(string, map->name) == 0)
			return map->value;
	
	return (isdigit(*string)) ? atoi(string) : def;
}

static struct conf_var *find(const char *name, struct conf_var *start)
{
	if (!start)
		start = conf_head;
	while (start && xstrcasecmp(start->name, name) != 0)
		start = start->next;
	return start;
}

int conf_get_int(const char *name, int def)
{
	struct conf_var *r;
	return (r = find(name, 0)) == 0 ? def : atoi(r->val);
}

int conf_get_maped(const char *name, int def, struct conf_map *map)
{
	struct conf_var *r;
	return (r = find(name, 0)) == 0 ? def : convert(r->val, map, def);
}

const char *conf_get_string(const char *name, const char *def)
{
	struct conf_var *r;
	return (r = find(name, 0)) == 0 ? def : r->val;
}

int conf_get_bool(const char *name, int def)
{
	static struct conf_map map[] = {
		{ "true", 1 },
		{ "t", 1 },
		{ "yes", 1 },
		{ "y", 1 },
		{ "on", 1 },
		{ "false", 0 },
		{ "f", 0 },
		{ "no", 0 },
		{ "n", 0 },
		{ "off", 0 },
#ifndef STRICT_CONFIG
		{ "yap", 1 },
		{ "ya", 1 },
		{ "ok", 1 },
		{ "okey", 1 },
		{ "okay", 1 },
		{ "nope", 0 },
		{ "oh,no", 0 },
		{ "never", 0 },
		{ "always", 1 },
#endif
	};
	return conf_get_maped(name, def, map);
}

int conf_find_int(const char *name, int v)
{
	struct conf_var *r = 0;
	
	while ((r = find(name, r)))
		if (atoi(r->val) == v)
			return 1;
	return 0;
}

int conf_find_string(const char *name, const char *v)
{
	struct conf_var *r = 0;
	
	while ((r = find(name, r)))
		if (xstrcasecmp(r->val, v) == 0)
			return 1;
	return 0;
}

char *conf_file = CONFIG_NAME;

void conf_read()
{
	FILE *f;
	char line[2000];
	
	f = fopen(conf_file, "r");
	if (!f)
		return;
	while (fgets(line, sizeof(line), f)) {
		char *p = line, *p2;
		struct conf_var *cv;
		
		while (isspace(*p))
			p++;
		if (*p == 0 || *p == '#')
			continue;
		for (p2 = p; *p2 && !isspace(*p2); p2++) { }
		if (*p2) {
			*p2++ = 0;
			while (isspace(*p2))
				p2++;
		}
		
		cv = xmalloc(sizeof(*cv));
		cv->name = xstrdup(p);
		
		if (*p2) {
			p = p2;
			while (strchr("\n#", *p2) == 0)
				p2++;
			*p2 = 0;
			while (isspace(*--p2))
				*p2 = 0;
			cv->val = xstrdup(p);
		} else
			cv->val = xstrdup("");
		cv->next = conf_head;
		conf_head = cv;
	}
}
